/* 
 *    Pixel Perfect collision detection demo
 *    by Neil Chinnery
 */



#include <stdlib.h>
#include <stdio.h>
#include <sys/farptr.h>
#include <sys/segments.h>
#include <sys/movedata.h>
#include "allegro.h"
#include "sprites.h" /* Data file Index - contains some sprites and a pallete */
#define MAX_SPRITES 10

DATAFILE *data;
BITMAP *s;

struct spr_mask { /* Sprite Mask */
 int bb_height; /* Bounding-box height of sprite */
 int bb_width; /* Bounding-box width of sprite */
 unsigned long int sp_mask[32]; /* The 32 entries of the 32 bit Sprite Mask */
} spr_mask[MAX_SPRITES];


int x1=20, y1=120; /* X and Y co-ords of Sprite 1 */
int x2=150, y2=150; /* X and Y co-ors of Sprite 2 */
int sp1=1, sp2=2; /* Sprite 1 & 2 positions in datafile */

void init_sp_bb() /* initialise data in Sprite mask to be zeros */
{
 int l,c;

 for (l=0; l<(MAX_SPRITES+1); l++)
 {
  spr_mask[l].bb_height=0;
  spr_mask[l].bb_width=0;
  for (c=0; c<32; c++) spr_mask[l].sp_mask[c]=0;
 }
}

void mk_spr_mask(BITMAP *s3, int spr_no)
{
 int bb_height=0; /* Bounding box height of sprite */
 int bb_width=0; /* Bounding box width of sprite */
 int x1, y1; /* Used to span through pixels within the sprite */
 int p; /* Holds return value from getpixel() */

 for (y1=0; y1<32; y1++) /* Now go through each pixel of the sprite */
 {
  for (x1=0; x1<32; x1++)       /* in turn */
  {
   p=getpixel(s3,x1,y1);         /* Find out what colour the pixel is */
   if (p>0)                     /* if the colour is non-zero */
    {
     if (y1>bb_height) bb_height=y1;  /* if > than current bounding box width, increase bounding box width */
     if (x1>bb_width) bb_width=x1;    /* if > than current bounding box heigth, increase bounding box height */
     spr_mask[spr_no].sp_mask[y1]+=0x0001<<(31-x1); /* Set corresponding bit to be 'on' in the mask*/
    }
  }
 }
 spr_mask[spr_no].bb_height=bb_height; /* Set the final bounding box height and width */
 spr_mask[spr_no].bb_width=bb_width;
}

/* this routine does normal bounding-box collision detection */

int check_bbcollision(int spr1, int spr1x, int spr1y, int spr2, int spr2x, int spr2y)
 {
 if ((spr1x>spr2x+spr_mask[spr2].bb_width) || (spr2x > spr1x+spr_mask[spr1].bb_width) ||
     (spr1y>spr2y+spr_mask[spr2].bb_height) || (spr2y> spr1y+spr_mask[spr1].bb_height))
  {
     return 0;
  }
  else
  {
   return 1;
  }
}

int check_ppcollision(int spr1, int spr1x, int spr1y, int spr2, int spr2x, int spr2y)
{
  int dx1, dx2, dy1, dy2;

 /* first do normal bounding-box collision */

 if ((spr1x>spr2x+spr_mask[spr2].bb_width) || (spr2x > spr1x+spr_mask[spr1].bb_width) ||
     (spr1y>spr2y+spr_mask[spr2].bb_height) || (spr2y> spr1y+spr_mask[spr1].bb_height))
  {
     return 0;    /* No collision */
  }
  else   /* if bounding box reports collision do pixel-perfect check*/
  {
   if (spr1x>spr2x) {
     dx1=0;            //don't need to shift sprite 1
     dx2=spr1x-spr2x;  //we need to shift sprite 2 left
   }
   else {
     dx1=spr2x-spr1x;  //we need to shift sprite 1 left
     dx2=0;            //don't need to shift sprite 2
   }
   if (spr1y>spr2y) {
     dy1=0;            //we don't need to skip any rows on sprite 1
     dy2=spr1y-spr2y;  //we need to skip this many rows on sprite 2
   }
   else {
     dy1=(spr2y-spr1y)-1;  //we need to skip this many rows on sprite 1
     dy2=0;           // we don't need to skip any rows on sprite 2
   }
   while((dy1<spr_mask[spr1].bb_height)&&(dy2<=spr_mask[spr2].bb_height)) {
     if ((spr_mask[spr1].sp_mask[dy1]<<dx1)&(spr_mask[spr2].sp_mask[dy2]<<dx2))
       return 1; /* a true collision */
     dy1++;
     dy2++;
   }
   return 0; /* no collision */
 }
}


void mve_spr()
{
 if (key[KEY_UP]) y1--;         /* move sprite 1 around the screen */
 if (key[KEY_DOWN]) y1++;
 if (key[KEY_RIGHT]) x1++;
 if (key[KEY_LEFT]) x1--;
 if (key[KEY_1])
 {
  sp1++;             /* change bitmap of sprite 1 */
  clear_keybuf();
 }
 if (key[KEY_2])
 {
  sp2++;            /* change bitmap of sprite 2 */
  clear_keybuf();
 }
 if (y1<0) y1=0;        /* keep sprite 1 within the limits of the screen */
 if (y1>460) y1=460;
 if (x1<0) x1=0;
 if (x1>620) x1=620;
 if (sp1>7) sp1=1;
 if (sp2>7) sp2=1;

}

void drw_spr()
{
 char msg[80];
 clear_to_color(s,0);
 draw_sprite (s, data[sp1].dat, x1, y1);
 draw_sprite (s, data[sp2].dat, x2, y2);
 sprintf(msg, "Sprite (1)=%d", sp1);
 textout(s,font,msg, 10,10,255);
 sprintf(msg, "Bounding Box Height=%d", spr_mask[sp1].bb_height);
 textout(s,font,msg, 10,20,255);
 sprintf(msg, "Bounding Box Width=%d", spr_mask[sp1].bb_width);
 textout(s,font,msg, 10,30,255);
 sprintf(msg, "Sprite (2)=%d", sp2);
 textout(s,font,msg, 10,50,255);
 sprintf(msg, "Bounding Box Height=%d", spr_mask[sp2].bb_height);
 textout(s,font,msg, 10,60,255);
 sprintf(msg, "Bounding Box Width=%d", spr_mask[sp2].bb_width);
 textout(s,font,msg, 10,70,255);

 /* here's where we test for collision!
    Normally you would only use check_ppcollision, because the first part of
    this function does bounding-box collision-detection.  I have used
    check_bbcollision as well, to demonstrate that bounding-box is not very
    accurate and can return collisions when, in fact, none have occured */


 if (check_bbcollision(sp1,x1,y1,sp2,x2,y2)) /* Just Bounding Box Check */
    textout(s,font, "Bding Box Intersect - Col Detected!!",10,100,255);

 if (check_ppcollision(sp1,x1,y1,sp2,x2,y2)) /* Bounding Box and Pixel-Precision */
    textout(s,font, "Pixel Intersect - TRUE Collision!!",10,110,255);
 vsync();
 blit(s,screen,0,0,0,0,320,240);
}


void main()
{
 int c;
 char msg[80];


   init_sp_bb(); /* Initialise bounding-box structure */
   /* Initialise Allegro Units */
   allegro_init();
   install_keyboard(); 
   install_timer();

   data=load_datafile("sprites.dat"); /* Load the datafile */
   
   clear_keybuf(); /* Clear the keyboard buffer */
   
   /* Create bitmap to allow double-buffering */
   s=create_bitmap(640,480);

   clear_to_color(s,0);



   /* Create Bounding Boxes & Bit-Mask for the seven sprites in the datafile */

  for (c=1; c<8; c++) mk_spr_mask(data[c].dat, c);

  /* Set the Graphics mode */

   set_gfx_mode(GFX_AUTODETECT, 320, 240, 0, 0);
   set_pallete(data[PAL1].dat);
   do
   {
   mve_spr();  /* Move the sprites */
   drw_spr(); /*Display the sprites */
   }while(!key[KEY_ESC]);       /* Until the user presses esc */
   clear_keybuf();
   exit(0);
}
